from math import degrees

from FreeCAD import Vector


def place_shapes(shapes, placement):
    for shape in shapes:
        place_shape(shape, placement)


def place_shape(shape, placement):
    """Apply a placement to a given shape.

    See Part::TopoShapePy Class Reference:
        https://www.freecadweb.org/api/db/d71/classPart_1_1TopoShapePy.html

    :param part: A part.
    :type part: Part::TopoShapePy
    :param placement: Placement to apply to part
    :type placement: FreeCAD.Placement
    """
    rotation = placement.Rotation
    shape.rotate(Vector(), rotation.Axis, degrees(rotation.Angle))

    translation = placement.Base
    shape.translate(translation)



"""Contains cut list task panel widgets.
"""
from .copy_cut_list_to_clipboard_task_panel import \
    CopyCutListToClipboardTaskPanel
from .save_cut_list_as_csv_task_panel import SaveCutListAsCsvTaskPanel



def find_edges_connected_to_vertex(vertex, edges):
    vertex_edges = []
    for e in edges:
        for v in e.Vertexes:
            if v.isSame(vertex):
                vertex_edges.append(e)
    return vertex_edges



"""Exposes functions for the following topological data types or "shapes":
    * Compound - A group of any type of topological object.
    * Compsolid - A composite solid is a set of solids,
                  connected by their faces.
                  It expands the notions of WIRE and SHELL to solids.
    * Solid - A part of space limited by shells. It is three dimensional.
    * Shell - A set of faces connected by their edges.
              A shell can be open or closed.
    * Face - In 2D it is part of a plane; in 3D it is part of a surface.
             Its geometry is constrained (trimmed) by contours.
             It is two dimensional.
    * Wire - A set of edges connected by their vertices.
             It can be an open or closed contour,
             depending on whether the edges are linked or not.
    * Edge - A topological element corresponding to a restrained curve.
             An edge is generally limited by vertices.
             It has one dimension.
    * Vertex - A topological element corresponding to a point.
               It has zero dimension.

Shape is  generic term covering all of the above.

See Also:
    https://wiki.freecadweb.org/Topological_data_scripting
"""
from .place_shape import place_shape, place_shapes
from .move_parts import move_parts



import os

import FreeCADGui
from FreeCAD import Console
from osecore.app.cut_list import write_dict_list_to_csv
from PySide import QtGui

from .cut_list_task_panel_base import CutListTaskPanelBase


class SaveCutListAsCsvTaskPanel(CutListTaskPanelBase):

    def __init__(self, cut_list_table_rows, columns, note=None):
        title = 'Save Cut List as CSV'
        super(SaveCutListAsCsvTaskPanel, self).__init__(
            title, cut_list_table_rows, columns, note)
        self.cut_list_table_rows = cut_list_table_rows
        self.columns = columns

        path_to_default_file = find_available_path_to_default_file()
        self.path_to_default_file = path_to_default_file
        self.save_file_name = self.path_to_default_file

        row = QtGui.QHBoxLayout()
        self.label = QtGui.QLabel(self.form)
        self.label.setObjectName('filenameLabel')
        self.set_label_text()
        row.addWidget(self.label)
        file_select_button = QtGui.QPushButton('Select Other File')
        file_select_button.clicked.connect(self.handle_file_select)
        row.addWidget(file_select_button)
        self.layout.addLayout(row)

    def handle_file_select(self):
        csv_filter = 'csv(*.csv)'
        save_file_name = QtGui.QFileDialog.getSaveFileName(
            self.form,
            'Select File',
            self.path_to_default_file,
            csv_filter)[0]
        if save_file_name != '':
            self.save_file_name = save_file_name
            self.set_label_text()

    def set_label_text(self):
        filename = os.path.basename(self.save_file_name)
        last_dir = os.path.split(os.path.dirname(self.save_file_name))[-1]
        display_text = os.path.join('...', last_dir, filename)
        self.label.setText('<b>File:</b> ' + display_text)

    def accept(self):
        """
        Executed upon clicking "OK" button in FreeCAD Tasks panel.
        """
        Console.PrintMessage(
            'Saving cut list as "{}".\n'.format(self.save_file_name))
        write_dict_list_to_csv(
            self.cut_list_table_rows,
            self.columns,
            self.save_file_name)
        FreeCADGui.Control.closeDialog()


def find_available_path_to_default_file():
    default_filename_template = 'cutlist{}.csv'
    path_to_desktop = os.path.expanduser('~/Desktop')
    path_to_default_file_template = os.path.join(
        path_to_desktop, default_filename_template)
    path_to_default_file = path_to_default_file_template.format('')
    if os.path.exists(path_to_default_file):
        i = 1
        while os.path.exists(path_to_default_file):
            path_to_default_file = path_to_default_file_template.format(i)
            i += 1
    return path_to_default_file



from PySide import QtCore, QtGui


class CutListTaskPanelBase:
    """Base class for cut list task panels.

    Handles building cut list table and initializing layout.
    """

    def __init__(self, title, cut_list_table_rows, columns, note=None):
        self.cut_list_table_rows = cut_list_table_rows
        self.columns = columns

        self.form = QtGui.QWidget()
        self.form.setWindowTitle(title)
        self.layout = QtGui.QVBoxLayout(self.form)
        row = QtGui.QHBoxLayout()
        table = build_cut_list_table_widget(cut_list_table_rows, columns)
        self.layout.addLayout(row)
        row.addWidget(table)

        if note is not None:
            row2 = QtGui.QHBoxLayout()
            note_label = QtGui.QLabel(self.form)
            note_label.setObjectName('note_label')
            note_text = '<b>Note:</b> {}'.format(note)
            note_label.setText(note_text)
            note_label.setWordWrap(True)
            row2.addWidget(note_label)
            self.layout.addLayout(row2)

    def accept(self):
        """
        Executed upon clicking "OK" file_select_button in FreeCAD Tasks panel.
        """
        raise NotImplementedError('Sub class must implement accept() method.')


def build_cut_list_table_widget(cut_list_table_rows, columns):
    num_rows = len(cut_list_table_rows)
    num_columns = len(columns)
    table = QtGui.QTableWidget(num_rows, num_columns)
    table.setHorizontalHeaderLabels(columns)

    for row_index, row in enumerate(cut_list_table_rows):
        for column_index, value in enumerate(row.values()):
            item = QtGui.QTableWidgetItem(value)
            # make table cell not editable
            item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEditable)
            item.setToolTip(value)
            table.setItem(row_index, column_index, item)

    header = table.horizontalHeader()
    header.setResizeMode(QtGui.QHeaderView.Stretch)
    return table



from osecore.app.cut_list import (convert_cut_list_to_ordered_dicts,
                                  merge_cut_list_items_by_length)

from .cut_list_task_type import CutListTaskType
from .task_panel import (CopyCutListToClipboardTaskPanel,
                         SaveCutListAsCsvTaskPanel)


class CutListTaskPanelFactory:
    """Creates Cut List Task Panel objects.
    """

    def __init__(self,
                 cut_list,
                 merge_cut_list=False,
                 note=None):
        self.cut_list = _transform_cut_list(cut_list, merge_cut_list)
        self.note = note

    def create(self, cut_list_task_type):
        columns = ['Quantity', 'Description', 'Length']
        if cut_list_task_type == CutListTaskType.CopyToClipboard:
            return CopyCutListToClipboardTaskPanel(
                self.cut_list, columns, self.note)
        elif cut_list_task_type == CutListTaskType.SaveToFile:
            return SaveCutListAsCsvTaskPanel(
                self.cut_list, columns, self.note)
        else:
            message_template = 'Unrecognized cut list task type "{}".'
            raise ValueError(message_template.format(cut_list_task_type))


def _transform_cut_list(cut_list, should_merge_cut_list):
    ordered_dict_cut_list = convert_cut_list_to_ordered_dicts(cut_list)
    if should_merge_cut_list:
        return merge_cut_list_items_by_length(ordered_dict_cut_list)
    else:
        return ordered_dict_cut_list



from math import degrees

import Part
from FreeCAD import Console, Rotation, Vector
from osecore.app.shape.edge import (find_edges_connected_to_vertex,
                                    is_edge_parallel_to_x_axis,
                                    is_edge_parallel_to_y_axis,
                                    is_edge_parallel_to_z_axis)


def move_parts(parts,
               placement,
               origin_translation_offset,
               reference_dimensions,
               rotation=Rotation()):
    translation_offset = _get_translation_offset(
        reference_dimensions, rotation, origin_translation_offset)
    for part in parts:
        part.translate(translation_offset)
        placement_rotation = placement.Rotation
        part.rotate(translation_offset, placement_rotation.Axis,
                    degrees(placement_rotation.Angle))
        part.translate(placement.Base)


def _get_translation_offset(reference_dimensions,
                            rotation,
                            origin_translation_offset):
    reference_box = Part.makeBox(*reference_dimensions)
    reference_box.rotate(Vector(0, 0, 0), rotation.Axis,
                         degrees(rotation.Angle))
    first_vertex = reference_box.Vertexes[0]
    first_vertex_edges = find_edges_connected_to_vertex(
        first_vertex, reference_box.Edges)
    num_edges = len(first_vertex_edges)
    if num_edges != 3:
        message = 'Found {} edges'.format(num_edges)
        message += ' connected to cube vertex instead of 3.\n'
        Console.PrintWarning(message)
    x = y = z = 1.0
    for e in first_vertex_edges:
        if is_edge_parallel_to_x_axis(e):
            x = e.Length
        elif is_edge_parallel_to_y_axis(e):
            y = e.Length
        elif is_edge_parallel_to_z_axis(e):
            z = e.Length
        else:
            def round_vec(v):
                return Vector(round(v.x), round(v.y), round(v.z))
            Console.PrintWarning(
                'Edge: {} not parallel to x, y, or z axes\n'.format(
                    [round_vec(v.Point) for v in e.Vertexes]))

    return Vector(
        x * origin_translation_offset.x,
        y * origin_translation_offset.y,
        z * origin_translation_offset.z
    )



import csv

from FreeCAD import Console


def write_dict_list_to_csv(dict_list, columns, filename):
    try:
        with open(filename, 'w') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=columns)
            writer.writeheader()
            for row in dict_list:
                writer.writerow(row)
    except IOError as e:
        Console.PrintError(str(e))



"""Exposes functions to create cut list task panels.
"""
from .create_copy_cut_list_to_clipboard_task_panel import \
    create_copy_cut_list_to_clipboard_task_panel
from .create_save_cut_list_to_file_task_panel import \
    create_save_cut_list_to_file_task_panel



import FreeCADGui
from FreeCAD import Console
from osecore.app.cut_list import (convert_dict_list_to_wiki_list_markup,
                                  convert_dict_list_to_wiki_table_markup)
from PySide import QtGui

from .cut_list_task_panel_base import CutListTaskPanelBase


class CopyCutListToClipboardTaskPanel(CutListTaskPanelBase):

    def __init__(self, cut_list_table_rows, columns, note=None):
        title = 'Copy Cut List to Clipboard'
        super(CopyCutListToClipboardTaskPanel, self).__init__(
            title, cut_list_table_rows, columns, note)

        # Row 1 - Markup Format options
        row1 = QtGui.QVBoxLayout()

        # Table option
        self.table_option = QtGui.QRadioButton('Table', self.form)
        self.table_option.setChecked(True)
        row1.addWidget(self.table_option)

        # List option
        self.list_option = QtGui.QRadioButton('List', self.form)
        row1.addWidget(self.list_option)

        self.layout.addLayout(row1)

        # ----------------------------------

        # Row 2
        row2 = QtGui.QHBoxLayout()

        # Copy Status Label
        self.copyStatusLabel = QtGui.QLabel(self.form)
        self.copyStatusLabel.setObjectName('copyStatusLabel')
        self.copyStatusLabel.setText('<b>Status:</b> Not Copied')
        row2.addWidget(self.copyStatusLabel)

        # Copy to Clipboard button
        copy_to_clipboard_button = QtGui.QPushButton(
            'Copy to Clipboard')
        copy_to_clipboard_button.clicked.connect(self.handle_copy_to_clipboard)
        row2.addWidget(copy_to_clipboard_button)

        self.layout.addLayout(row2)

    def handle_copy_to_clipboard(self):
        clipboard = QtGui.QApplication.clipboard()
        table_markup = ''
        if self.table_option.isChecked():
            table_markup = convert_dict_list_to_wiki_table_markup(
                self.cut_list_table_rows, self.columns)
        elif self.list_option.isChecked():
            table_markup = convert_dict_list_to_wiki_list_markup(
                self.cut_list_table_rows)
        else:
            raise ValueError('Table or list option must be checked.')
        clipboard.setText(table_markup)
        self.copyStatusLabel.setText('<b>Status:</b> Copied!')
        Console.PrintMessage('Copied table markup to clipboard.\n')

    def accept(self):
        """
        Executed upon clicking "OK" button in FreeCAD Tasks panel.
        """
        self.handle_copy_to_clipboard()
        FreeCADGui.Control.closeDialog()



from functools import reduce


def convert_dict_list_to_wiki_list_markup(dict_list):
    return reduce(row_to_list_item, dict_list, '')


def row_to_list_item(table_markup, row):
    list_item_template = u"* ['''{}'''] {} \u2014 {}\n"
    return table_markup + list_item_template.format(
        row['Quantity'], row['Description'], row['Length'])



def convert_dict_list_to_wiki_table_markup(dict_list, columns):
    """
    See:
        https://www.mediawiki.org/wiki/Help:Tables#Wiki_table_markup_summary

    :param dict_list: list of dictionaries
    :type dict_list: List(Dict)
    :param columns: Columns of table
    :type columns: list of strings
    :return: wiki table markup
    :rtype: string
    """
    table = '{| class="wikitable"\n'
    row_separator = '|----\n'
    for column in columns:
        table += '!|{}\n'.format(column)
    table += row_separator
    for row in dict_list:
        for value in row.values():
            table += '|{}\n'.format(value)
        table += row_separator
    return table + '|}\n'



from .cut_list_task_panel_factory import CutListTaskPanelFactory
from .cut_list_task_type import CutListTaskType


def create_save_cut_list_to_file_task_panel(
        cut_list_items,
        merge_cut_list_items_by_length=False,
        note=None):
    factory = CutListTaskPanelFactory(
        cut_list_items,
        merge_cut_list=merge_cut_list_items_by_length,
        note=note)
    return factory.create(CutListTaskType.SaveToFile)



from FreeCAD import Console


def find_selection_object_by_type(selection_objects, object_type):
    return _find_object_by_type(
        selection_objects,
        'selection objects',
        object_type,
        lambda selection_object: _is_selection_object_type_of(
            selection_object, object_type)
    )


def find_face_in_selection_object(selection_object):
    return _find_sub_object_by_shape_type(selection_object, ShapeType.FACE)


def find_edge_in_selection_object(selection_object):
    return _find_sub_object_by_shape_type(selection_object, ShapeType.EDGE)


def find_vertex_in_selection_object(selection_object):
    return _find_sub_object_by_shape_type(selection_object, ShapeType.VERTEX)


def _find_sub_object_by_shape_type(selection_object, shape_type):
    if selection_object is None:
        return None
    sub_objects = selection_object.SubObjects
    return _find_object_by_type(
        sub_objects,
        'sub objects',
        shape_type,
        lambda sub_object: _is_sub_object_type_of(sub_object, shape_type)
    )


def _find_object_by_type(objects, subject, object_type, filter_predicate):
    potential_objects = list(filter(filter_predicate, objects))
    num_matches = len(potential_objects)
    if num_matches == 0:
        return None
    if num_matches > 1:
        _print_warning_message(num_matches, subject, object_type)
    return potential_objects[0]


class ShapeType:
    """Shape Type Enumeration meant to mirror PartGui::DimSelections::ShapeType
    in FreeCAD C++ code.

    See:
        https://www.freecadweb.org/api/da/d26/classPartGui_1_1DimSelections.html
    """
    VERTEX = 'Vertex'
    EDGE = 'Edge'
    FACE = 'Face'


def _is_sub_object_type_of(sub_object, shape_type):
    return sub_object.ShapeType == shape_type


def _is_selection_object_type_of(selection_object, object_type):
    return (
        selection_object.Object.TypeId == 'Part::FeaturePython' and
        selection_object.Object.Proxy.Type == object_type
    )


def _print_warning_message(num_matches, subject, object_type):
    message_template = '{} {} matching type "{}" found in selection.'
    message_template += ' Returning first match.'
    message = message_template.format(num_matches, subject, object_type)
    Console.PrintWarning(message)
